"""
This model describes a free-floating vehicle-sharing system.
This package contains the classes that extend the basemodel and implement the model's features.
More information on the :ref:`FF_VS` page.
"""
