"""
This model describes a station-based vehicle-sharing system.
This package contains the classes that extend the basemodel and implement the model's features.
More information on the :ref:`SB_VS` page.
"""
